@@ -1,3 +1,11 @@
+;;   Copyright (c) Nicola Mometto, Rich Hickey & contributors.
+;;   The use and distribution terms for this software are covered by the
+;;   Eclipse Public License 1.0 (http://opensource.org/licenses/eclipse-1.0.php)
+;;   which can be found in the file epl-v10.html at the root of this distribution.
+;;   By using this software in any fashion, you are agreeing to be bound by
+;;   the terms of this license.
+;;   You must not remove this notice, or any other, from this software.
+
 (ns ^{:doc "A clojure reader in clojure"
       :author "Bronsa"}
   clojure.tools.reader
@@ -10,7 +18,8 @@
   (:import (clojure.lang PersistentHashSet IMeta
                          RT Symbol Reflector Var IObj
                          PersistentVector IRecord Namespace)
-           java.lang.reflect.Constructor))
+           java.lang.reflect.Constructor
+           java.util.regex.Pattern))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; helpers
@@ -31,8 +40,7 @@
   [rdr initch]
   (if-not initch
     (reader-error rdr "EOF while reading")
-    (loop [sb (doto (StringBuilder.) (.append initch))
-           ch (read-char rdr)]
+    (loop [sb (StringBuilder.) ch initch]
       (if (or (whitespace? ch)
               (macro-terminating? ch)
               (nil? ch))
@@ -68,13 +76,13 @@
        (loop [i offset uc 0]
          (if (== i l)
            (char uc)
-           (let [d (Character/digit (char (nth token i)) (int base))]
+           (let [d (Character/digit (int (nth token i)) (int base))]
              (if (== d -1)
                (throw (IllegalArgumentException. (str "Invalid digit: " (nth token i))))
                (recur (inc i) (long (+ d (* uc base))))))))))
 
   ([rdr initch base length exact?]
-     (loop [i 1 uc (Character/digit (char initch) (int base))]
+     (loop [i 1 uc (Character/digit (int initch) (int base))]
        (if (== uc -1)
          (throw (IllegalArgumentException. (str "Invalid digit: " initch)))
          (if-not (== i length)
@@ -86,7 +94,7 @@
                  (throw (IllegalArgumentException.
                          (str "Invalid character length: " i ", should be: " length)))
                  (char uc))
-               (let [d (Character/digit (char ch) (int base))]
+               (let [d (Character/digit (int ch) (int base))]
                  (read-char rdr)
                  (if (== d -1)
                    (throw (IllegalArgumentException. (str "Invalid digit: " ch)))
@@ -100,7 +108,10 @@
   [rdr backslash]
   (let [ch (read-char rdr)]
     (if-not (nil? ch)
-      (let [token (read-token rdr ch)
+      (let [token (if (or (macro-terminating? ch)
+                          (whitespace? ch))
+                    (str ch)
+                    (read-token rdr ch))
             token-len (count token)]
         (cond
 
@@ -146,48 +157,73 @@
         (if (identical? delim (char ch))
           (persistent! a)
           (if-let [macrofn (macros ch)]
-            (let [mret (macrofn rdr ch)]
+            (let [mret (log-source-unread rdr
+                         (macrofn rdr ch))]
               (recur (if-not (identical? mret rdr) (conj! a mret) a)))
             (let [o (read (doto rdr (unread ch)) true nil recursive?)]
               (recur (if-not (identical? o rdr) (conj! a o) a)))))
         (reader-error rdr "EOF while reading"
                       (when first-line
-                        (str ", starting at line" first-line)))))))
+                        (str ", starting at line " first-line)))))))
 
 (defn- read-list
   [rdr _]
-  (let [[line column] (when (indexing-reader? rdr)
-                        [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-        the-list (read-delimited \) rdr true)]
-    (if (empty? the-list)
-      '()
-      (with-meta (clojure.lang.PersistentList/create the-list)
-        (when line
-          {:line line :column column})))))
+  (let [[start-line start-column] (when (indexing-reader? rdr)
+                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
+        the-list (read-delimited \) rdr true)
+        [end-line end-column] (when (indexing-reader? rdr)
+                                [(get-line-number rdr) (int (get-column-number rdr))])]
+    (with-meta (if (empty? the-list)
+                 '()
+                 (clojure.lang.PersistentList/create the-list))
+      (when start-line
+        (merge
+         (when-let [file (get-file-name rdr)]
+           {:file file})
+         {:line start-line
+          :column start-column
+          :end-line end-line
+          :end-column end-column})))))
 
 (defn- read-vector
   [rdr _]
-  (let [[line column] (when (indexing-reader? rdr)
-                        [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-        the-vector (read-delimited \] rdr true)]
+  (let [[start-line start-column] (when (indexing-reader? rdr)
+                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
+        the-vector (read-delimited \] rdr true)
+        [end-line end-column] (when (indexing-reader? rdr)
+                                [(get-line-number rdr) (int (get-column-number rdr))])]
     (with-meta the-vector
-      (when line
-        {:line line :column column}))))
+      (when start-line
+        (merge
+         (when-let [file (get-file-name rdr)]
+           {:file file})
+         {:line start-line
+          :column start-column
+          :end-line end-line
+          :end-column end-column})))))
 
 (defn- read-map
   [rdr _]
-  (let [[line column] (when (indexing-reader? rdr)
-                        [(get-line-number rdr) (int (dec (get-column-number rdr)))])
+  (let [[start-line start-column] (when (indexing-reader? rdr)
+                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
         the-map (read-delimited \} rdr true)
-        map-count (count the-map)]
+        map-count (count the-map)
+        [end-line end-column] (when (indexing-reader? rdr)
+                                [(get-line-number rdr) (int (get-column-number rdr))])]
     (when (odd? map-count)
       (reader-error rdr "Map literal must contain an even number of forms"))
     (with-meta
       (if (zero? map-count)
         {}
         (RT/map (to-array the-map)))
-      (when line
-        {:line line :column column}))))
+      (when start-line
+        (merge
+         (when-let [file (get-file-name rdr)]
+           {:file file})
+         {:line start-line
+          :column start-column
+          :end-line end-line
+          :end-column end-column})))))
 
 (defn- read-number
   [reader initch]
@@ -211,11 +247,11 @@
       \b "\b"
       \f "\f"
       \u (let [ch (read-char rdr)]
-           (if (== -1 (Character/digit (char ch) 16))
+           (if (== -1 (Character/digit (int ch) 16))
              (reader-error rdr "Invalid unicode escape: \\u" ch)
              (read-unicode-char rdr ch 16 4 true)))
       \x (let [ch (read-char rdr)]
-           (if (== -1 (Character/digit (char ch) 16))
+           (if (== -1 (Character/digit (int ch) 16))
              (reader-error rdr "Invalid unicode escape: \\x" ch)
              (read-unicode-char rdr ch 16 2 true)))
       (if (numeric? ch)
@@ -238,9 +274,9 @@
 
 (defn- read-symbol
   [rdr initch]
-  (when-let [token (read-token rdr initch)]
-    (let [[line column] (when (indexing-reader? rdr)
-                          [(get-line-number rdr) (int (dec (get-column-number rdr)))])]
+  (let [[line column] (when (indexing-reader? rdr)
+                        [(get-line-number rdr) (int (dec (get-column-number rdr)))])]
+    (when-let [token (read-token rdr initch)]
       (case token
 
         ;; special symbols
@@ -255,11 +291,25 @@
         (or (when-let [p (parse-symbol token)]
               (with-meta (symbol (p 0) (p 1))
                 (when line
-                  {:line line :column column})))
+                  (merge
+                   (when-let [file (get-file-name rdr)]
+                     {:file file})
+                   {:line line
+                    :column column
+                    :end-line (get-line-number rdr)
+                    :end-column (int (inc (get-column-number rdr)))}))))
             (reader-error rdr "Invalid token: " token))))))
 
+(def ^:dynamic *alias-map*
+  "Map from ns alias to ns, if non-nil, it will be used to resolve read-time
+   ns aliases instead of (ns-aliases *ns*).
+
+   Defaults to nil"
+  nil)
+
 (defn- resolve-ns [sym]
-  (or ((ns-aliases *ns*) sym)
+  (or ((or *alias-map*
+           (ns-aliases *ns*)) sym)
       (find-ns sym)))
 
 (defn- read-keyword
@@ -268,7 +318,7 @@
     (if-not (whitespace? ch)
       (let [token (read-token reader ch)
             s (parse-symbol token)]
-        (if (and s (== -1 (.indexOf token "::")))
+        (if s
           (let [^String ns (s 0)
                 ^String name (s 1)]
             (if (identical? \: (nth token 0))
@@ -289,31 +339,43 @@
 
 (defn- read-meta
   [rdr _]
-  (let [[line column] (when (indexing-reader? rdr)
-                        [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-        m (desugar-meta (read rdr true nil true))]
-    (when-not (map? m)
-      (reader-error rdr "Metadata must be Symbol, Keyword, String or Map"))
-    (let [o (read rdr true nil true)]
-      (if (instance? IMeta o)
-        (let [m (if (and line
-                         (seq? o))
-                  (assoc m :line line
-                           :column column)
-                  m)]
-          (if (instance? IObj o)
-            (with-meta o (merge (meta o) m))
-            (reset-meta! o m)))
-        (reader-error rdr "Metadata can only be applied to IMetas")))))
+  (log-source rdr
+    (let [[line column] (when (indexing-reader? rdr)
+                          [(get-line-number rdr) (int (dec (get-column-number rdr)))])
+          m (desugar-meta (read rdr true nil true))]
+      (when-not (map? m)
+        (reader-error rdr "Metadata must be Symbol, Keyword, String or Map"))
+      (let [o (read rdr true nil true)]
+        (if (instance? IMeta o)
+          (let [m (if (and line (seq? o))
+                    (assoc m :line line :column column)
+                    m)]
+            (if (instance? IObj o)
+              (with-meta o (merge (meta o) m))
+              (reset-meta! o m)))
+          (reader-error rdr "Metadata can only be applied to IMetas"))))))
 
 (defn- read-set
   [rdr _]
-  (PersistentHashSet/createWithCheck (read-delimited \} rdr true)))
+  (let [[start-line start-column] (when (indexing-reader? rdr)
+                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
+        the-set (PersistentHashSet/createWithCheck (read-delimited \} rdr true))
+        [end-line end-column] (when (indexing-reader? rdr)
+                                [(get-line-number rdr) (int (get-column-number rdr))])]
+    (with-meta the-set
+      (when start-line
+        (merge
+         (when-let [file (get-file-name rdr)]
+           {:file file})
+         {:line start-line
+          :column start-column
+          :end-line end-line
+          :end-column end-column})))))
 
 (defn- read-discard
   [rdr _]
-  (read rdr true nil true)
-  rdr)
+  (doto rdr
+    (read true nil true)))
 
 (def ^:private ^:dynamic arg-env)
 
@@ -330,16 +392,15 @@
           rargs (rseq arg-env)
           args (if rargs
                  (let [higharg (key (first rargs))]
-                   (if (pos? higharg)
-                     (let [args (loop [i 1 args (transient [])]
-                                  (if (> i higharg)
-                                    (persistent! args)
-                                    (recur (inc i) (conj! args (or (get arg-env i)
-                                                                   (garg i))))))
-                           args (if (arg-env -1)
-                                  (conj args '& (arg-env -1))
-                                  args)]
-                       args)))
+                   (let [args (loop [i 1 args (transient [])]
+                                (if (> i higharg)
+                                  (persistent! args)
+                                  (recur (inc i) (conj! args (or (get arg-env i)
+                                                                 (garg i))))))
+                         args (if (arg-env -1)
+                                (conj args '& (arg-env -1))
+                                args)]
+                     args))
                  [])]
       (list 'fn* args form))))
 
@@ -360,50 +421,26 @@
     (read-symbol rdr pct)
     (let [ch (peek-char rdr)]
       (cond
-        (or (whitespace? ch)
-            (macro-terminating? ch)
-            (nil? ch))
-        (register-arg 1)
+       (or (whitespace? ch)
+           (macro-terminating? ch)
+           (nil? ch))
+       (register-arg 1)
 
-        (identical? ch \&)
-        (do (read-char rdr)
-            (register-arg -1))
+       (identical? ch \&)
+       (do (read-char rdr)
+           (register-arg -1))
 
-        :else
-        (let [n (read rdr true nil true)]
-          (if-not (integer? n)
-            (throw (IllegalStateException. "Arg literal must be %, %& or %integer"))
-            (register-arg n)))))))
+       :else
+       (let [n (read rdr true nil true)]
+         (if-not (integer? n)
+           (throw (IllegalStateException. "Arg literal must be %, %& or %integer"))
+           (register-arg n)))))))
 
 (defn- read-eval
   [rdr _]
   (when-not *read-eval*
     (reader-error rdr "#= not allowed when *read-eval* is false"))
-  (let [o (read rdr true nil true)]
-    (if (symbol? o)
-      (RT/classForName (str ^Symbol o))
-      (if (list? o)
-        (let [fs (first o)
-              o (rest o)
-              fs-name (name fs)]
-          (cond
-           (= fs 'var) (let [vs (first o)]
-                         (RT/var (namespace vs) (name vs)))
-           (.endsWith fs-name ".")
-           (let [args (to-array o)]
-             (-> fs-name (subs 0 (dec (count fs-name)))
-                 RT/classForName (Reflector/invokeConstructor args)))
-
-           (Compiler/namesStaticMember fs)
-           (let [args (to-array o)]
-             (Reflector/invokeStaticMethod (namespace fs) fs-name args))
-
-           :else
-           (let [v (Compiler/maybeResolveIn *ns* fs)]
-             (if (var? v)
-               (apply v o)
-               (reader-error rdr "Can't resolve " fs)))))
-        (throw (IllegalArgumentException. "Unsupported #= form"))))))
+  (eval (read rdr true nil true)))
 
 (def ^:private ^:dynamic gensym-env nil)
 
@@ -440,8 +477,8 @@
     (if s
       (let [e (first s)]
         (recur (next s) (-> key-vals
-                            (conj! (key e))
-                            (conj! (val e)))))
+                          (conj! (key e))
+                          (conj! (val e)))))
       (seq (persistent! key-vals)))))
 
 (defn- register-gensym [sym]
@@ -454,7 +491,7 @@
         (set! gensym-env (assoc gensym-env sym gs))
         gs)))
 
-(defn- resolve-symbol [s]
+(defn ^:dynamic resolve-symbol [s]
   (if (pos? (.indexOf (name s) "."))
     s
     (if-let [ns-str (namespace s)]
@@ -472,18 +509,23 @@
 
 (defn- add-meta [form ret]
   (if (and (instance? IObj form)
-           (dissoc (meta form) :line :column))
+           (seq (dissoc (meta form) :line :column :end-line :end-column :file :source)))
     (list 'clojure.core/with-meta ret (syntax-quote* (meta form)))
     ret))
 
 (defn- syntax-quote-coll [type coll]
-  (let [res (list 'clojure.core/seq
+  (let [res (list 'clojure.core/sequence
                   (cons 'clojure.core/concat
                         (expand-list coll)))]
     (if type
       (list 'clojure.core/apply type res)
       res)))
 
+(defn map-func [coll]
+  (if (>= (count (:val coll)) 16)
+    'clojure.core/hash-map
+    'clojure.core/array-map))
+
 (defn- syntax-quote* [form]
   (->>
    (cond
@@ -494,7 +536,7 @@
           (if (namespace form)
             (let [maybe-class ((ns-map *ns*)
                                (symbol (namespace form)))]
-              (if (class? class)
+              (if (class? maybe-class)
                 (symbol (.getName ^Class maybe-class) (name form))
                 (resolve-symbol form)))
             (let [sym (name form)]
@@ -515,21 +557,26 @@
 
     (coll? form)
     (cond
+
      (instance? IRecord form) form
-     (map? form) (syntax-quote-coll 'clojure.core/hash-map (flatten-map form))
-     (vector? form) (syntax-quote-coll 'clojure.core/vector form)
+     (map? form) (syntax-quote-coll (map-func form) (flatten-map form))
+     (vector? form) (list 'clojure.core/vec (syntax-quote-coll nil form))
      (set? form) (syntax-quote-coll 'clojure.core/hash-set form)
      (or (seq? form) (list? form))
      (let [seq (seq form)]
        (if seq
          (syntax-quote-coll nil seq)
          '(clojure.core/list)))
+
      :else (throw (UnsupportedOperationException. "Unknown Collection type")))
 
     (or (keyword? form)
         (number? form)
         (char? form)
-        (string? form))
+        (string? form)
+        (nil? form)
+        (instance? Boolean form)
+        (instance? Pattern form))
     form
 
     :else (list 'quote form))
@@ -539,7 +586,7 @@
   [rdr backquote]
   (binding [gensym-env {}]
     (-> (read rdr true nil true)
-        syntax-quote*)))
+      syntax-quote*)))
 
 (defn- macros [ch]
   (case ch
@@ -589,16 +636,17 @@
                              \{ [\} :extended]
                              nil)]
       (let [entries (to-array (read-delimited end-ch rdr true))
+            numargs (count entries)
             all-ctors (.getConstructors class)
             ctors-num (count all-ctors)]
         (case form
           :short
           (loop [i 0]
-            (if (> i ctors-num)
+            (if (>= i ctors-num)
               (reader-error rdr "Unexpected number of constructor arguments to " (str class)
-                            ": got" (count entries))
+                            ": got" numargs)
               (if (== (count (.getParameterTypes ^Constructor (aget all-ctors i)))
-                      ctors-num)
+                      numargs)
                 (Reflector/invokeConstructor class entries)
                 (recur (inc i)))))
           :extended
@@ -686,27 +734,40 @@
      (when (= :unknown *read-eval*)
        (reader-error "Reading disallowed - *read-eval* bound to :unknown"))
      (try
-       (let [ch (read-char reader)]
-         (cond
-          (whitespace? ch) (read reader eof-error? sentinel recursive?)
-          (nil? ch) (if eof-error? (reader-error reader "EOF") sentinel)
-          (number-literal? reader ch) (read-number reader ch)
-          (comment-prefix? ch) (read (read-comment reader ch) eof-error? sentinel recursive?)
-          :else (let [f (macros ch)]
-                  (if f
-                    (let [res (f reader ch)]
-                      (if (identical? res reader)
-                        (read reader eof-error? sentinel recursive?)
-                        res))
-                    (read-symbol reader ch)))))
+       (loop []
+         (log-source reader
+           (let [ch (read-char reader)]
+             (cond
+              (whitespace? ch) (recur)
+              (nil? ch) (if eof-error? (reader-error reader "EOF") sentinel)
+              (number-literal? reader ch) (read-number reader ch)
+              (comment-prefix? ch) (do (read-comment reader) (recur))
+              :else (let [f (macros ch)]
+                      (if f
+                        (let [res (f reader ch)]
+                          (if (identical? res reader)
+                            (recur)
+                            res))
+                        (read-symbol reader ch)))))))
        (catch Exception e
          (if (ex-info? e)
-           (throw e)
+           (let [d (ex-data e)]
+             (if (= :reader-exception (:type d))
+               (throw e)
+               (throw (ex-info (.getMessage e)
+                               (merge {:type :reader-exception}
+                                      d
+                                      (if (indexing-reader? reader)
+                                        {:line   (get-line-number reader)
+                                         :column (get-column-number reader)
+                                         :file   (get-file-name reader)}))
+                               e))))
            (throw (ex-info (.getMessage e)
                            (merge {:type :reader-exception}
                                   (if (indexing-reader? reader)
-                                    {:line (get-line-number reader)
-                                     :column (get-column-number reader)}))
+                                    {:line   (get-line-number reader)
+                                     :column (get-column-number reader)
+                                     :file   (get-file-name reader)}))
                            e)))))))
 
 (defn read-string
@@ -725,7 +786,8 @@
   (when (and s (not (identical? s "")))
     (read (string-push-back-reader s) true nil false)))
 
-(defmacro syntax-quote [form]
+(defmacro syntax-quote
   "Macro equivalent to the syntax-quote reader macro (`)."
+  [form]
   (binding [gensym-env {}]
     (syntax-quote* form)))
